from syntax import *
from cfg import Cfg
from iteration import *
from opsem import eval_bexp, eval_aexp

one = AECst(1)
zero = AECst(0)
x = AEVar("x")
dead_if=ICond(BEPlain(one),ISet("x",one),ISet("y",zero))
infinite_loop=ILoop(BEPlain(one),ISet("x",AEBop(Bop.ADD,x,one)))
prog_with_dead_code=ISeq(dead_if,infinite_loop)

class NoInfo(Transfer[frozenset[bool]]):
    """Almost trivial static analysis
 
    We only have two elements in our lattice, the empty set and the singleton {True}.
    The latter is propagated from initial state to any potentially reachable state.
    In the end, the only nodes still at bottom are the one which are guarded by
    always-false tests.
    """
    type S = frozenset[bool]
    def bottom(self) -> S: return frozenset()
    def init_state(self) -> S: return frozenset([True])
    def join(self, s1:S, s2:S) -> S: return s1 | s2
    def included(self, s1:S, s2:S) -> bool: return s1.issubset(s2)
    def tr_skip(self,s:S): return s
    def tr_set(self, s:S, v:str, e: ArithExpr) -> S: return s
    def tr_test(self,s:S, c:BoolExpr) -> S:
        if eval_bexp({},c) is False:
            return frozenset()
        else:
            return s
    def tr_err(self, s:S, e:Expr) -> S:
        if isinstance(e,ArithExpr):
            if eval_aexp({},e) is not None:
                return frozenset()
            else:
                return s
        else:
            if isinstance(e,BoolExpr):
                if eval_bexp({},e) is not None:
                    return frozenset()
                else:
                    return s
            else: assert False

cfg = Cfg(prog_with_dead_code)
print(f"Cfg is {cfg}")
res = fixpoint_iteration(NoInfo(),cfg)
unreachable = [node for node, state in res.items() if state == frozenset()]
for node in unreachable:
    print(f"{node} is unreachable")
